//********************************************************
// uWatch
// Common Calculator Functions
// Version 1.2
// Last Update: 22nd May 08
// Written for the Microchip C30 Compiler
// Target Device: PIC24FJ64GA004 (44pin)
// Copyright(c) 2008 David L. Jones
// http://www.calcwatch.com
// EMAIL: david@alternatezone.com

// NOTE: This code is designed to be inserted inline into the uWatch-Main.C function
//       Most variables are passed as globals and are defined in uWatch-Main.C
//********************************************************

/*********************************************************
This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**********************************************************/

extern void Conversions(void);
extern void HexEntry(void);
extern void UpdateTRIGdisplay(void);
extern void UpdateXregDisplay(void);
extern void UpdateYregDisplay(void);
extern void UpdateDisplayRegs(void);
extern void CompleteXreg(void);
extern void ResetFlags(void);
extern void SwitchMenuOff(void);
extern double Factorial(double num);
extern void StoreRecall(void);
extern void SignKey(void);
extern void KeyRecord(void);
extern void KeyReplay(void);
extern void CalcMenu(void);
extern double Deg(double num);
// These routines are common to both the RPN and Algebraic routines

//converts randians to degrees
double Deg(double num)
{
	return((180/PI)*num);
}

//converts degrees to radians
double Rad(double num)
{
	return((PI*num)/180);
}

//***********************************
// we cheat here and update the trig display seperately from the other menu items as it is dynamic.
void UpdateTRIGdisplay(void)
{
	//display DEG/RAD mode
	if (CurrentMenu==1)
		{
		if (DegreesMode==TRUE) 
			UpdateLCDline2(" ARC  HYP (DEG)");
		else
			UpdateLCDline2(" ARC  HYP (RAD)");
		}

}

//***********************************
//Converts both X&Y regs from real numbers into display strings and then displays them
//Also displays the current algebraic operator
void UpdateDisplayRegs(void)
{
	char c;
	unsigned int i;
	UpdateXregDisplay();
	UpdateYregDisplay();

	if (RPNmode==TRUE)
		{
		UpdateLCDline1(DisplayYreg);
		UpdateLCDline2(DisplayXreg);
		return;		//nothing more to do in RPN mode
		}

	//the rest of this code is for the ALG mode only
	switch (OperatorXY)
		{
		case OPRplus:  c='+'; break;
		case OPRminus: c='-'; break;
		case OPRmult:  c='x'; break;
		case OPRdiv:   c='/'; break;
		case OPRparra: c='|'; break;
		case OPRnPr:   c='P'; break;
		case OPRnCr:   c='C'; break;
		case OPRxpwry: c='^'; break;
		case OPRrtop:  c='>'; break;
		case OPRptor:  c='<'; break;
		default: c=' ';break;
		}
	for(i=0; i<=15; i++)
		if (DisplayYreg[i]==NULL) DisplayYreg[i]=' ';	//replace NULL with space
	DisplayYreg[15]=c;				//display the operator character
	DisplayYreg[16]=NULL;

	if (OperatorXY1!=0) 
		{
		DisplayYreg[13]='(';				//display the operator character	
		DisplayYreg[14]='1';
		}
	if (OperatorXY2!=0) DisplayYreg[14]='2';
	if (OperatorXY3!=0) DisplayYreg[14]='3';
	if (OperatorXY4!=0) DisplayYreg[14]='4';
	if (OperatorXY5!=0) DisplayYreg[14]='5';
	if (OperatorXY6!=0) DisplayYreg[14]='6'; 
	
	UpdateLCDline1(DisplayYreg);
	UpdateLCDline2(DisplayXreg);
}

//*************************************************
// all the 6 functions for menu item 0
// common to both RPN and ALG modes
void Menu0(int MenuItem)
{
	switch(MenuItem)
		{
		case 1:		//1/x key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			Xreg=1/Xreg;			
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		
		case 2:		//X^2  key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			Xreg=Xreg*Xreg;			
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		
		case 3:		//SQRT key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			if (Xreg<0) Xreg=42; else Xreg=sqrt(Xreg);			//HHGTTG
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
	
		case 4:		//LOG key presssed
			{
			CompleteXreg();		//enter value on stack if needed
			Xreg=log10(Xreg);			
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
	
		case 5:		//10^X key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			Xreg=pow(10,Xreg);			
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
	
		case 6:		//LOGe key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			Xreg=log(Xreg);			
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		}
}

//*************************************************
// all the 6 functions for menu item 1
// common to both RPN and ALG modes
void Menu1(int MenuItem)
{
	switch(MenuItem)
		{
		case 1:			// SIN key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			if (InverseKey==TRUE) 
				{
				Xreg=asin(Xreg);
				if (DegreesMode==TRUE) Xreg=Deg(Xreg);
				}
			if (HYPkey==TRUE) 
				{
				Xreg=sinh(Xreg);
				}
			if ((InverseKey==FALSE) && (HYPkey==FALSE)) 
				{
				if (DegreesMode==TRUE) Xreg=Rad(Xreg);
				Xreg=sin(Xreg);			
				}
			InverseKey=FALSE;		//reset the function flags
			HYPkey=FALSE;
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		case 2:			// COS key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			if (InverseKey==TRUE) 
				{
				Xreg=acos(Xreg);
				if (DegreesMode==TRUE) Xreg=Deg(Xreg);
				}
			if (HYPkey==TRUE) 
				{
				Xreg=cosh(Xreg);
				}
			if ((InverseKey==FALSE) && (HYPkey==FALSE)) 
				{
				if (DegreesMode==TRUE) Xreg=Rad(Xreg);
				Xreg=cos(Xreg);			
				}
			InverseKey=FALSE;		//reset the function flags
			HYPkey=FALSE;
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		case 3:			// TAN key pressed
			{
			CompleteXreg();		//enter value on stack if needed
			if (InverseKey==TRUE) 
				{
				Xreg=atan(Xreg);
				if (DegreesMode==TRUE) Xreg=Deg(Xreg);
				}
			if (HYPkey==TRUE) 
				{
				Xreg=tanh(Xreg);
				}
			if ((InverseKey==FALSE) && (HYPkey==FALSE)) 
				{
				if (DegreesMode==TRUE) Xreg=Rad(Xreg);
				Xreg=tan(Xreg);			
				}
			InverseKey=FALSE;		//reset the function flags
			HYPkey=FALSE;
			UpdateDisplayRegs();	//update display again
			SwitchMenuOff();
			break;
			}
		case 4:			//INV key pressed
			{
			InverseKey=TRUE;
			UpdateLCDline2(" Inverse");
			break;
			}
		case 5:			//HYP key pressed
			{
			HYPkey=TRUE;
			UpdateLCDline2(" Hyperbolic");
			break;
			}
		case 6:			// DEG/RAD key
			{
			if (DegreesMode==TRUE)
				DegreesMode=FALSE;
			else DegreesMode=TRUE;
			UpdateTRIGdisplay();
			}
		}

}

//***********************************
//perform the factorial operation and return the result
double Factorial(double num)
{
	double temp;

	temp=1;
	while(num>=1)		//perform the factorial algorithm
		{
		temp=temp*num;
		num=num-1;
		}
	return(temp);
}

//***********************************
// Checks to see if the Xreg has been ENTERed onto the stack
// If not then it it adds the value to the stack
// This routien is needed because the displayed Xreg value is not stored in the real Xreg variable yet
void CompleteXreg(void)
{
	if (ValueEntered==FALSE)	//if ENTER has not been pressed then do it for them
		{
		Xreg=atof(DisplayXreg);		//convert display string to number
		ResetFlags();
		ValueEntered=TRUE;		//value is now entered on the stack, so we must flag this.
		}
}

//***********************************
// Converts the Xreg double value into a string for the DisplayXreg
void UpdateXregDisplay(void)
{
	strcpy(DisplayXreg,"                ");		//blank the string first
	sprintf(DisplayXreg,"%.9g",Xreg);
	return;
}

//***********************************
// Converts the Yreg double value into a string for the DisplayYreg
void UpdateYregDisplay(void)
{
	strcpy(DisplayYreg,"                ");		//blank the string first
	sprintf(DisplayYreg,"%.8g",Yreg);
	return;
}

//**********************************
//switches the menu off and restores the Xreg/Yreg display
void SwitchMenuOff(void)
{
	MenuMode=FALSE;
	UpdateLCDline1(DisplayYreg);
	UpdateLCDline2(DisplayXreg);
}

//***********************************
//resets all the flags after a calculation, ready for the next number to be entered by the user
void ResetFlags(void)
{
	ValueEntered=TRUE;
	DecimalIncluded=FALSE;
	MinusIncluded=FALSE;
	MinusIncludedInExponent=FALSE;
	ExponentIncluded=FALSE;
	EnableXregOverwrite=FALSE;		//calling function must overwirte this if needed
}

//***********************************
//processes the store/recall function
void StoreRecall(void)
{
	char KeyPress2;
	int num;

		//this operation discards any value currently in the display register
	UpdateLCDline1("RECALL REG 0-9 ?");		//ask for a register number
	do KeyPress2=KeyScan(); while(KeyPress2==0);	//get the users response		
	if (KeyPress2==KeyRCL)						//check for 2nd press of RCl key to enter STO mode
		{										// do the STO function
		UpdateLCDline1(" STORE REG 0-9 ?");		//ask for a register number
		do KeyPress2=KeyScan(); while(KeyPress2==0);	//get the users response		
		switch(KeyPress2)
			{
			case Key0: num=0; break;
			case Key1: num=1; break;
			case Key2: num=2; break;
			case Key3: num=3; break;
			case Key4: num=4; break;
			case Key5: num=5; break;
			case Key6: num=6; break;
			case Key7: num=7; break;
			case Key8: num=8; break;
			case Key9: num=9; break;
			default: num=10; 			//this is the escape code, user pressed something else other than 0-9
			}					
		if(num!=10)
			{
			Sreg[num]=Xreg;			//store the Xreg value in the appropriate Sreg
			UpdateLCDline1("  VALUE STORED  ");
			DelayMs(100);			//small visual delay
			}
		}
	else				//do the RCL function
		{
		switch(KeyPress2)
			{
			case Key0: num=0; break;
			case Key1: num=1; break;
			case Key2: num=2; break;
			case Key3: num=3; break;
			case Key4: num=4; break;
			case Key5: num=5; break;
			case Key6: num=6; break;
			case Key7: num=7; break;
			case Key8: num=8; break;
			case Key9: num=9; break;
			default: num=10; 			//this is the escape code, user pressed something else other than 0-9
			}					
		if(num!=10)
			{
			Xreg=Sreg[num];			//restore the Xreg value from the Sreg
			ValueEntered=TRUE;
			UpdateLCDline1(" VALUE RECALLED ");
			DelayMs(100);			//small visual delay
			}
		}
}

//**********************************
// processes the sign key
void SignKey(void)
{
	int num,c;
	if (MenuMode==FALSE)			//this key is not used in MENU mode			
	{
		if (ExponentIncluded==TRUE)		//do this if there is an exponent already
			{
			num=0;			//assume cannot find 'e'
			for (c=0; c<strlen(DisplayXreg); c++)	//search Xreg for 'e'
				if (DisplayXreg[c]=='e') num=c;		//found the exponent
			//num now holds the position of 'e'
			if (MinusIncludedInExponent==TRUE)		//minus already exists so lets remove it
				{
				for(c=num+1;c<strlen(DisplayXreg);c++)
					DisplayXreg[c]=DisplayXreg[c+1];	//shift all characters one space left to remove existing minus sign
				MinusIncludedInExponent=FALSE;
				}
			else		//minus does not exist yet so lets add one
				{
	//							for(c=num+1;c>0;c--)
	//								DisplayXreg[c+1]=DisplayXreg[c];	//shift all characters one space right
				DisplayXreg[num+1]='-';
				DisplayXreg[num+2]=NULL;
				MinusIncludedInExponent=TRUE;
				}
			UpdateLCDline2(DisplayXreg);		//update the display
			}
		else							//do this if there is only a mantissa
			{
			if (ValueEntered==TRUE) 	//if this is the first digit pressed
				{
				if (EnableXregOverwrite==FALSE)	//check to see if we don't have to overwrite the Xreg
				strcpy(DisplayXreg,"");		//clear (overwrite)what was in the Xreg
				EnableXregOverwrite=FALSE;	//disable overwriting the Xreg for future key presses
				}
	
			if (MinusIncluded==TRUE)
				{
				MinusIncluded=FALSE;
				for(c=0;c<strlen(DisplayXreg);c++)
					DisplayXreg[c]=DisplayXreg[c+1];	//shift all characters one space left to remove existing minus sign
				}
			else
				{
				MinusIncluded=TRUE;
				for(c=strlen(DisplayXreg)+1;c>0;c--)
					DisplayXreg[c]=DisplayXreg[c-1];	//shift all characters one space right
				DisplayXreg[0]='-';						//include minus sign
				}
			}
		ValueEntered=FALSE;
		UpdateLCDline2(DisplayXreg);		//update the display
	}
}

void KeyRecord(void)
{
	char Key;
	unsigned int num,c;
	char s[MaxLCDdigits+1];

	UpdateLCDline1("Keystroke Rec");
	UpdateLCDline2("Program#(00-59)");
	do Key=KeyScan(); while(Key==0);		//now get a two digit number from the user
	if (Key==KeyClear) return;
	num=ReturnNumber(Key)*10;				//this is the 10's digit
	do Key=KeyScan(); while(Key==0);
	num=num+ReturnNumber(Key);
	if (num>59) num=59;
	MemPointer=(num*1024)+16;		//set the memory pointer to after the header info
	sprintf(s,"Program %2i",num);
	for(c=0;c<15;c++) I2CmemoryWRITE((MemPointer-16)+c,s[c]);
	ProgRec=TRUE;						//switch on keystroke programming mode
	UpdateLCDline1(s);
	UpdateLCDline2("Rec Keys Now...");
	DelayMs(1000);
}

void KeyReplay(void)
{
	char Key;
	unsigned int num,c;
	char s[MaxLCDdigits+1];

	UpdateLCDline1("Keystroke Play");
	UpdateLCDline2("Program#(00-59)");
	do Key=KeyScan(); while(Key==0);		//now get a two digit number from the user
	if (Key==KeyClear) return;
	num=ReturnNumber(Key)*10;				//this is the 10's digit
	do Key=KeyScan(); while(Key==0);
	num=num+ReturnNumber(Key);
	if (num>59) num=59;
	MemPointer=(num*1024)+16;		//set the memory pointer to after the header info
	for(c=0;c<15;c++) s[c]=I2CmemoryREAD((MemPointer-16)+c);
	s[16]=NULL;
	ProgPlay=TRUE;						//switch on keystroke programming mode
	UpdateLCDline1(s);
	UpdateLCDline2("Running...");
	DelayMs(1000);
}

void CalcMenu(void)
{
	if (MenuMode==TRUE)		//menu is already on so switch menu screens
		{
		CurrentMenu++;				
		if (CurrentMenu>=MaxRPNmenuItems) 
			{		//end of menu so display stats
			CurrentMenu=0;	//wrap back to start of menu
			}
		UpdateLCDline1(CalcMenuLine1[CurrentMenu]);
		UpdateLCDline2(CalcMenuLine2[CurrentMenu]);
		}
	else			//switch on the menu
		{
		MenuMode=TRUE;
		UpdateLCDline1(CalcMenuLine1[CurrentMenu]);
		UpdateLCDline2(CalcMenuLine2[CurrentMenu]);
		}	
	UpdateTRIGdisplay(); //not very elegant, but gotta do it somewhere
}


void RtoP(void)
{
	double temp;
	temp=sqrt((Yreg*Yreg)+(Xreg*Xreg));
	Xreg=atan(Xreg/Yreg);
	if (DegreesMode==TRUE) Xreg=Deg(Xreg);
	Yreg=temp;
}

void PtoR(void)
{
	double temp;
	if (DegreesMode==TRUE) temp=Yreg*cos(Rad(Xreg));
		else temp=Yreg*cos(Xreg);
		
	if (DegreesMode==TRUE) Xreg=Yreg*sin(Rad(Xreg));
		else Xreg=Yreg*sin(Xreg);
	Yreg=temp;
}

//**************************************
// enter a HEX digit A-F onto the current Xreg string
void HexEntry(void)
{
	int KeyPress2;
	char digit[2];
	UpdateLCDline1(" F1=A F2=B F3=C ");
	UpdateLCDline2(" F4=D F5=E F6=F ");
	do KeyPress2=KeyScan(); while(KeyPress2==0);
	switch(KeyPress2)
		{
		case Key7: {strcpy(digit,"A"); break;}
		case Key8: {strcpy(digit,"B"); break;}
		case Key9: {strcpy(digit,"C"); break;}
		case Key4: {strcpy(digit,"D"); break;}
		case Key5: {strcpy(digit,"E"); break;}
		case Key6: {strcpy(digit,"F"); break;}
		default:
			{
			UpdateLCDline1(DisplayYreg);
			UpdateLCDline2(DisplayXreg);
			return;			
			}
		}
		
	if (strlen(DisplayXreg)<MaxLCDdigits)
		{
		if (ValueEntered==TRUE) 	//if this is the first digit pressed
			{
			if (EnableXregOverwrite==FALSE)	//check to see if we don't have to overwrite the Xreg
			strcpy(DisplayXreg,"");		//clear (overwrite)what was in the Xreg
			EnableXregOverwrite=FALSE;	//disable overwriting the Xreg for future key presses
			}
		strcat(DisplayXreg,digit);
		ValueEntered=FALSE;
		MenuMode=FALSE;					//flag that a number has been entered
		UpdateLCDline1(DisplayYreg);
		UpdateLCDline2(DisplayXreg);
		}
}	

//**********************************
// display the conversions menu
void Conversions(void)
{
	unsigned int Key, KeyPress2;		//keypress variables
	char s[17];
	int Mode, Mode2;
	unsigned long int value;

	Mode=0;

	while(TRUE) 	//main keypress loop is infinite.
	{
		switch(Mode)
		{
			case 0: strcpy(s,"mils -> mm"); break;
			case 1: strcpy(s,"mm -> mils"); break;
			case 2: strcpy(s,"Deg F -> Deg C"); break;
			case 3: strcpy(s,"Deg C -> Deg F"); break;
			case 4: strcpy(s,"kg -> lb"); break;
			case 5: strcpy(s,"lb -> kg"); break;
			case 6: strcpy(s,"BASE N"); break;
		}

		UpdateLCDline1(s);
		UpdateLCDline2("CONV: +/- & ENT");

		do {			//wait for a keypress
			Key=KeyScan();
		} while(Key==0);

		ResetSleepTimer();		
		//start to process the keypress

		if (Key==KeyClear) return;		// clear key was pressed, exit  mode
		if (Key==KeyMode) return;		// mode key was pressed, exit  mode

		NextMode=FALSE;				//user pressed some key other than MODE, so ensure that when exit we go back to the time/date display
		ResetSleepTimer();

		if (Key==KeyPlus)				//user has pressed the PLUS key
			Mode++;
		if (Mode>5) Mode=0;

		if (Key==KeyMinus)				//user has pressed the MINUS key
			Mode--;
		if (Mode<0) Mode=5;

		if (Key==KeyEnter)				//user has pressed the ENTER key
			{
			switch(Mode)				
				{
					
				case 0:					//change date
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=Xreg*0.0254;
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}
				
				case 1:
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=Xreg*(1.0/0.0254);
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}
				
				case 2:					//clear the EEPROM contents
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=(Xreg-32.0)*5.0/9.0;
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}
				
				case 3:
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=Xreg*9.0/5.0+32.0;
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}

				case 4:
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=Xreg*2.2046226218487757;
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}

				case 5:
					{
					CompleteXreg();		//enter value on stack if needed
					Xreg=Xreg*0.45359237;
					UpdateDisplayRegs();	//update display again
					SwitchMenuOff();
					return;
					}
					
				case 6:					//BASE-N mode
					{
					Mode2=0;
					
					while(TRUE) 	//main keypress loop is infinite.
						{
							switch(Mode2)
							{
								case 0: strcpy(s,"Dec -> Hex"); break;
								case 1: strcpy(s,"Dec -> Bin"); break;
								case 2: strcpy(s,"Hex -> Dec"); break;
								case 3: strcpy(s,"Hex -> Bin"); break;
								case 4: strcpy(s,"Bin -> Dec"); break;
								case 5: strcpy(s,"Bin -> Hex"); break;
							}
					
							UpdateLCDline1(s);
							UpdateLCDline2("BASEN: +/- & ENT");
					
							do KeyPress2=KeyScan(); while(KeyPress2==0);
					
							ResetSleepTimer();		
							//start to process the keypress
					
							if (KeyPress2==KeyClear) return;	// clear key was pressed, exit setup mode
			
							if (KeyPress2==KeyPlus) Mode2++;	//user has pressed the PLUS key
							if (Mode2>5) Mode2=0;
					
							if (KeyPress2==KeyMinus) Mode2--;		//user has pressed the MINUS key
							if (Mode2<0) Mode2=5;
					
							if (KeyPress2==KeyEnter)				//user has pressed the ENTER key
								{
								switch(Mode2)				
									{
									case 0:			//decimal to hex
										{
											CompleteXreg();		//enter value on stack if needed
											value=0;
											do
												{
												
												}
											while(1);
											break;
										}
									}
								}
						}
					return;
					}

				}
			}
	}	//end of entire loop

}	
